//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include <stdio.h>
#include <algorithm>

#include "main.h"
#include "about.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
//---------------------------------------------------------------------------

/** \brief ccmpare function used with std::sort */
bool CompareFNADDR (const S_FNADDR s1, const S_FNADDR s2)
{
	return s1.address < s2.address;
}

/** \brief string hex to int taken from http://www.codeproject.com/string/hexstrtoint.asp */
int _httoi(const TCHAR *value)
{
  struct CHexMap
  {
    TCHAR chr;
    int value;
  };
  const int HexMapL = 16;
  CHexMap HexMap[HexMapL] =
  {
    {'0', 0}, {'1', 1},
    {'2', 2}, {'3', 3},
    {'4', 4}, {'5', 5},
    {'6', 6}, {'7', 7},
    {'8', 8}, {'9', 9},
    {'A', 10}, {'B', 11},
    {'C', 12}, {'D', 13},
    {'E', 14}, {'F', 15}
  };
  TCHAR *mstr = _tcsupr(_tcsdup(value));
  TCHAR *s = mstr;
  int result = 0;
  if (*s == '0' && *(s + 1) == 'X') s += 2;
  bool firsttime = true;
  while (*s != '\0')
  {
    bool found = false;
    for (int i = 0; i < HexMapL; i++)
    {
      if (*s == HexMap[i].chr)
      {
        if (!firsttime) result <<= 4;
        result |= HexMap[i].value;
        found = true;
        break;
      }
    }
    if (!found) break;
    s++;
    firsttime = false;
  }
  free(mstr);
  return result;
}

__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner)
{
	slFileLines=new TStringList;
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::About1Click(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::Exit1Click(TObject *Sender)
{
	this->Close();	
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::Openmapfile1Click(TObject *Sender)
{
	if(OpenDialog->Execute())
	{
		lblStatus->Caption = "Reading MAP...";
		lblStatus->Visible = true;
		lblStatus->Update();
		pbProgress->Position = 0;
		pbProgress->Visible = true;
		slFileLines->Clear();
		functions.clear();
		slFileLines->LoadFromFile(OpenDialog->FileName);
		for (int i=0; i<slFileLines->Count; i++)
		{
			if(!(i%256))
			{
				pbProgress->Position = pbProgress->Min + pbProgress->Max * (float)i/(float)slFileLines->Count;
				pbProgress->Update();
			}
			if(slFileLines->Strings[i].Pos(STR_ADDRESS))
			{
				AnsiString asAddress;
				struct S_FNADDR fnaddr;
				//parse remaining strings: address + function name
				while(++i < slFileLines->Count)
				{
					if(!(i%256))
					{
						pbProgress->Position = pbProgress->Max * (float)i/(float)slFileLines->Count;
						pbProgress->Update();
					}
					if(slFileLines->Strings[i].Pos(":"))
					{
						asAddress = slFileLines->Strings[i].SubString(slFileLines->Strings[i].Pos(":")+1, 8);
						fnaddr.address = _httoi(asAddress.c_str());
						if(fnaddr.address)
						{
							if(slFileLines->Strings[i].Pos("  "))
							{
								fnaddr.function = slFileLines->Strings[i].SubString(slFileLines->Strings[i].Pos("  ")+2,
									slFileLines->Strings[i].Length() - slFileLines->Strings[i].Pos("  "));
								if(fnaddr.function.Length())
									functions.push_back(fnaddr);
							}
						}
					}
				}
				//sort list
				std::sort (functions.begin(), functions.end(), CompareFNADDR);
				break;
			}
        }
	}
    pbProgress->Visible = false;
	lblStatus->Caption = IntToStr(functions.size()) + " function names loaded";
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
	delete slFileLines;
	slFileLines = NULL;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnSearchClick(TObject *Sender)
{
	if(!functions.size())
	{
		MessageBox(this->Handle, "MAP file not loaded or function list is empty",
			this->Caption.c_str(), MB_ICONEXCLAMATION);
		return;
    }
	if( _httoi(edAVaddress->Text.c_str()) < (_httoi(edImageBase->Text.c_str())+_httoi(edCodeBase->Text.c_str())) )
	{
		MessageBox(this->Handle, "AV address must be greater than Image Base plus Code Base",
			this->Caption.c_str(), MB_ICONEXCLAMATION);
		return;
    }
	lblStatus->Caption = "Searching...";
	lblStatus->Visible = true;
	lblStatus->Update();
	pbProgress->Position = pbProgress->Min;
	pbProgress->Visible = true;
	pbProgress->Update();

	unsigned int iAVaddress = _httoi(edAVaddress->Text.c_str()) -
						_httoi(edImageBase->Text.c_str()) -
						_httoi(edCodeBase->Text.c_str());
	if(iAVaddress < functions[0].address)
	{
		MessageBox(this->Handle, "AV address minus Image Base minus Code Base is smaller than least function address found in map file",
			this->Caption.c_str(), MB_ICONEXCLAMATION);
	}
	else
	{
		for(int i=functions.size()-1; i>=0; i--)
		{
			if(!(i%256))
			{
				pbProgress->Position = pbProgress->Min + pbProgress->Max * (float)i/(float)functions.size();
				pbProgress->Update();
			}
			if(iAVaddress >= functions[i].address)
			{
				edFunction->Text = functions[i].function;
				edDiff->Text = IntToStr(iAVaddress - functions[i].address);
				if(iAVaddress - functions[i].address > 10000)
				{
					MessageBox(this->Handle, "Difference beetween AV address and largest function address in MAP is large.\nInput data probably is wrong, exception can't be traced.",
						this->Caption.c_str(), MB_ICONEXCLAMATION);
				}
				break;
			}
		}
	}
	pbProgress->Visible = false;
	lblStatus->Visible = false;
}
//---------------------------------------------------------------------------

